import click
import os
import re
import subprocess
import sys

from .utils import die, error, header, important, info

# ModuleBuilder is a straightforward, simple wrapper around:
#
#   mbs-manager build_module_locally
#
# Except that it extensively filters the logging from that command
# to produce something that intended for a packager creating a module
# rather than a maintainer of an MBS installation.

class ModuleBuilder(object):
    def __init__(self, profile, modulemd, stream, local_builds=[]):
        self.profile = profile
        self.git_repo = self.find_git_repo()
        if self.git_repo:
            self.top_dir = self.git_repo
        else:
            self.top_dir = os.getcwd()

        if modulemd is not None:
            self.modulemd = modulemd
        else:
            self.modulemd = self.find_modulemd()

        if not self.modulemd.endswith(".yaml"):
            die("ModuleMD file '{}' does not have .yaml suffix")

        self.name = os.path.basename(self.modulemd)[:-5]
        if stream is not None:
            self.stream = stream
        else:
            self.stream = self.find_stream()

        self.local_builds = local_builds

    def find_git_repo(self):
        try:
            with open(os.devnull, 'w') as devnull:
                return subprocess.check_output(['git', 'rev-parse', '--show-toplevel'],
                                               stderr=devnull, encoding="UTF-8").strip()
        except subprocess.CalledProcessError:
            return None

    def find_modulemd(self):
        basename = os.path.basename(self.top_dir)
        modulemd = os.path.join(self.top_dir, basename + '.yaml')
        if not os.path.exists(modulemd):
            die("'{}' does not exist".format(modulemd))

        return modulemd

    def find_stream(self):
        if self.git_repo is None:
            return 'master'

        try:
            with open(os.devnull, 'w') as devnull:
                branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                                                  stderr=devnull, encoding="UTF-8").strip()
        except subprocess.CalledProcessError:
            branch = None

        if branch is None or branch == 'HEAD':
            die("Cannot determine current git branch")

        return branch

    def build(self):
        header('BUILDING MODULE')

        log_line_re = re.compile(r'.*?-\s*(\S+)\s-\s*(DEBUG|INFO|WARNING|ERROR)\s*-\s*(.*)')
        command_re = re.compile(r'Executing command: \[([^]]+)\](?:, stdout log: (\S+?))?(?:, stderr log: (\S+?))?$')
        build_id_re = re.compile(r'The user "[^"]+" submitted the build "([^"]+)"')

        # These define lines in the log output to ignore, in addition to all
        # messages at level DEBUG
        ignore_res = [re.compile(r) for r in [
            r'Calling complete',
            r'Calling done',
            r'Calling init',
            r'Calling wait',
            r'Cannot re-use',
            r'Connecting to koji',
            r'Found build',
            r'Hub not initialized',
            r'Machine arch setting',
            r'MockModuleBuilder initialized',
            r'<ModuleBuild ',
            r'No MBSConsumer found.  Shutting down\?',
            r'Saw relevant component build ',
            r'Scheduling faked event ',
            r'u"document: modulemd',
        ]]

        # These are lines that if we see in the log, we consider the build to have failed
        # (Possible we want to treat level=ERROR as fatal always)
        fail_res = [re.compile(r) for r in [
            r'Error while building artifact',
            r'Could not process message handler',
        ]]

        cmd = ['mbs-manager', 'build_module_locally',
               '--file', self.modulemd,
               '--stream', self.stream]
        for build_id in self.local_builds:
            cmd += ['--add-local-build', build_id]

        env = os.environ.copy()
        if self.profile.mbs_config_file is not None:
            env['MBS_CONFIG_FILE'] = self.profile.mbs_config_file
        if self.profile.mbs_config_section is not None:
            env['MBS_CONFIG_SECTION'] = self.profile.mbs_config_section
        process = subprocess.Popen(cmd,
                                   env=env,
                                   bufsize=1,
                                   universal_newlines=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT)

        output_dir = None
        log_path = None
        log_file = None
        log_lines = []
        failed = False

        while True:
            line = process.stdout.readline()
            if len(line) == 0:
                break

            if output_dir is None or not os.path.exists(output_dir):
                log_lines.append(line)
            else:
                if log_file is None:
                    log_file = open(log_path, 'w')
                    for l in log_lines:
                        log_file.write(l)
                log_file.write(line)
                log_file.flush()

            line = line.rstrip()
            m = log_line_re.match(line)
            if m is None:
                click.secho(line)
            else:
                logname = m.group(1)
                level = m.group(2)
                message = m.group(3)

                # Since https://pagure.io/fm-orchestrator/pull-request/1161
                # module-build-service messages are categorized as MBS.<subcategory>
                if logname != 'module_build_service' and not logname.startswith('MBS.'):
                    continue

                if level == 'DEBUG':
                    continue

                ignored = False
                for p in ignore_res:
                    if p.match(message) is not None:
                        ignored = True
                        break
                if ignored:
                    continue

                for p in fail_res:
                    if p.match(message) is not None:
                        failed = True
                        break

                if level == 'INFO':
                    m = build_id_re.match(message)
                    if m is not None:
                        nsvc = m.group(1)
                        parts = nsvc.split(":")
                        name = parts[0]
                        assert name == self.name
                        stream = parts[1]
                        assert stream == self.stream
                        version = parts[2]

                        output_dir = os.path.join(os.path.expanduser('~/modulebuild'),
                                                  'builds',
                                                  'module-{}-{}-{}'.format(name,
                                                                           stream,
                                                                           version))
                        log_path = os.path.join(output_dir, 'mbs.log')
                        important("name: {}".format(name))
                        important("stream: {}".format(stream))
                        important("version: {}".format(version))
                        important("modulemd: {}".format(self.modulemd))
                        important("log: {}".format(log_path))
                        click.echo()
                    else:
                        m = command_re.match(message)
                        if m is not None:
                            click.secho('running: ', fg='blue', bold=True, err=True, nl=False)
                            pieces = m.group(1).split(', ')
                            click.echo(' '.join(p[1:-1] for p in pieces))
                            if m.group(2) is not None and pieces[0] != "'mock'" and m.group(2) != '/dev/null':
                                click.secho('    stdout: ', fg='blue', err=True, nl=False)
                                click.echo(m.group(2))
                            if m.group(3) is not None:
                                click.secho('    stderr: ', fg='blue', err=True, nl=False)
                                click.echo(m.group(3))
                        else:
                            click.secho('info: ', fg='blue', bold=True, err=True, nl=False)
                            click.echo(message)

                elif level == 'WARNING':
                    click.secho('warning: ', fg='yellow', bold=True, err=True, nl=False)
                    click.echo(message)
                elif level == 'ERROR':
                    click.secho('error: ', fg='red', bold=True, err=True, nl=False)
                    click.echo(message)

        if log_file is not None:
            log_file.close()

        rv = process.wait()
        click.echo()

        if not failed and rv == 0:
            important('{}:{}:{} successfully built'.format(name, stream, version))
            info("log: {}".format(log_path))
        else:
            error("mbs-manager build_module_locally failed")
            error("log: {}".format(log_path))
            sys.exit(1)

        self.version = version




from collections import OrderedDict
import os
import re
import shutil

from .utils import check_call

import gi
gi.require_version('Modulemd', '2.0')
from gi.repository import Modulemd

from module_build_service.builder.utils import create_local_repo_from_koji_tag

from .utils import info, ModuleSpec
from .flatpak_builder import ModuleInfo
from .get_module_builds import get_module_builds

class Build(ModuleInfo):
    def yum_config(self):
        exclude = ','.join(self.mmd.get_rpm_filters())

        return """[{name}-{stream}]
name={name}-{stream}
baseurl=file://{path}
enabled=1
excludepkgs={exclude}
priority=10
# just treat modular packages as normal packages - don't expect modular metadata
module_hotfixes=true
""".format(name=self.name, stream=self.stream, path=self.path, exclude=exclude)

    def has_module_metadata(self):
        # Parsing the repomd.xml file would be cleaner and almost as simple,
        # but this matches how the dnf finds the modules file

        for f in os.listdir(os.path.join(self.path, 'repodata')):
            if "modules.yaml" in f:
                return True

        return False

class LocalBuild(Build):
    def __init__(self, path):
        mmd_path = os.path.join(path, 'modules.yaml')
        mmd = Modulemd.ModuleStream.read_file (mmd_path, False)
        mmd = mmd.upgrade(Modulemd.ModuleStreamVersionEnum.TWO)

        self.name = mmd.props.module_name
        self.stream = mmd.props.stream_name
        self.version = mmd.props.version

        self.path = path
        self.mmd = mmd

        self.rpms = [a + '.rpm' for a in mmd.get_rpm_artifacts()]

    def __repr__(self):
        return '<LocalBuild {name}:{stream}:{version}>'.format(**self.__dict__)

class KojiBuild(Build):
    def __init__(self, mmd, path, koji_tag, rpms):
        self.name = mmd.props.module_name
        self.stream = mmd.props.stream_name
        self.version = mmd.props.version

        self.path = path
        self.mmd = mmd

        self.rpms = rpms

        self.koji_tag = koji_tag

    def __repr__(self):
        return '<KojiBuild {name}:{stream}:{version}>'.format(**self.__dict__)


def get_module_info(module_name, stream, version=None, koji_config=None, koji_profile='koji'):
    builds = get_module_builds(module_name, stream, version=version,
                               koji_config=koji_config, koji_profile=koji_profile,
                               include_rpms=True)

    if len(builds) == 0:
        raise RuntimeError("No module builds found for {}"
                           .format(ModuleSpec(module_name, stream, version).to_str()))
    elif len(builds) > 1:
        raise RuntimeError("Multiple builds against different contexts found for {}"
                           .format(ModuleSpec(module_name, stream, version).to_str()))
    build = builds[0]

    modulemd_str = build['extra']['typeinfo']['module']['modulemd_str']
    mmd = Modulemd.ModuleStream.read_string(modulemd_str, False)
    # Make sure that we have the v2 'dependencies' format
    mmd = mmd.upgrade(Modulemd.ModuleStreamVersionEnum.TWO)

    rpms = ['{name}-{epochnum}:{version}-{release}.{arch}.rpm'.format(epochnum=rpm['epoch'] or 0, **rpm)
            for rpm in build['gmb_rpms']
            if rpm['arch'] in ('x86_64', 'noarch')]

    return mmd, build['extra']['typeinfo']['module']['content_koji_tag'], rpms


class ModuleLocator(object):
    class Config(object):
        pass

    def __init__(self, profile):
        self.profile = profile
        self.conf = ModuleLocator.Config()

        self.conf.koji_config = profile.koji_config
        self.conf.koji_profile = profile.koji_profile

        self.conf.cache_dir = os.path.expanduser('~/modulebuild/cache')
        self.conf.mock_resultsdir = os.path.expanduser('~/modulebuild/builds')

        self.local_build_ids = []
        self._local_build_info = None

        self._cached_remote_builds = {}

    def add_local_build(self, build_id):
        self.local_build_ids.append(build_id)
        self._local_build_info = None

    def get_local_build_info(self):
        if self._local_build_info is not None:
            return self._local_build_info

        if not self.local_build_ids:
            self._local_build_info = {}
            return self._local_build_info

        builds = []
        try:
            for d in os.listdir(self.conf.mock_resultsdir):
                m = re.match('^module-(.*)-([^-]*)-([0-9]+)$', d)
                if m:
                    builds.append((m.group(1), m.group(2), int(m.group(3)), d))
        except OSError:
            pass

        # Sort with the biggest version first
        builds.sort(key=lambda x: x[2], reverse=True)

        result = {}

        build_dir = self.conf.mock_resultsdir
        for build_id in self.local_build_ids:
            parts = build_id.split(':')
            if len(parts) < 1 or len(parts) > 3:
                raise RuntimeError(
                    'The local build "{0}" couldn\'t be be parsed into NAME[:STREAM[:VERSION]]'.format(build_id))

            name = parts[0]
            stream = parts[1] if len(parts) > 1 else None
            version = int(parts[2]) if len(parts) > 2 else None

            found_build = None
            for build in builds:
                if name != build[0]:
                    continue
                if stream is not None and stream != build[1]:
                    continue
                if version is not None and version != build[2]:
                    continue

                found_build = build
                break

            if not found_build:
                raise RuntimeError(
                    'The local build "{0}" couldn\'t be found in "{1}"'.format(build_id, self.conf.mock_resultsdir))

            local_build = LocalBuild(os.path.join(self.conf.mock_resultsdir, found_build[3], 'results'))

            if found_build[0] != local_build.name or \
               found_build[1] != local_build.stream or \
               found_build[2] != local_build.version:
                raise RuntimeError(
                    'Parsed metadata results for "{0}" don\'t match the directory name'.format(found_build[3]))
            result[(local_build.name, local_build.stream)] = local_build

        self._local_build_info = result
        return self._local_build_info

    def locate(self, name, stream, version=None):
        # FIXME: handle version
        key = (name, stream)
        if key in self.get_local_build_info():
            return self.get_local_build_info()[key]

        if key in self._cached_remote_builds:
            return self._cached_remote_builds[key]

        info("Querying Koji for information on %s:%s" % (name, stream))

        modulemd, koji_tag, rpms = get_module_info(name, stream,
                                                   version=version,
                                                   koji_config=self.conf.koji_config,
                                                   koji_profile=self.conf.koji_profile)

        path = os.path.join(self.conf.cache_dir, "modules", koji_tag)
        self._cached_remote_builds[key] = KojiBuild(modulemd, path, koji_tag, rpms)
        return self._cached_remote_builds[key]

    def ensure_downloaded(self, build):
        if os.path.exists(build.path):
            return

        # We want to re-use modules downloaded by MBS, but unfortunately, when
        # MBS downloads a module it doesn't include the module metadata, which
        # DNF wants to be able to do module dependencies. So we do a two step
        # process:
        #
        #  1) Use MBS code to download the module (which runs createrepo)
        #  2) Create another directory with a symlink to the MBS download,
        #     run createrepo *again* there, and then add in the module metadata
        #     with modifyrepo.
        #
        # (Adding module metadata directly to the MBS download directory would
        # break the module build, since MBS is expecting a bare repository
        # without module metadata.)
        #
        # Long term solution: Enable ODCS in Fedora infrastructure,
        #   stop downloading module builds

        koji_path = os.path.join(self.conf.cache_dir, "koji_tags", build.koji_tag)
        if not os.path.exists(koji_path):
            info("Downloading %s:%s to %s" % (build.name, build.stream, build.path))
            create_local_repo_from_koji_tag(self.conf, build.koji_tag, koji_path)

        os.makedirs(build.path)
        success = False
        try:
            os.symlink(os.path.join("../../koji_tags/", build.koji_tag),
                       os.path.join(build.path, "rpms"))
            check_call(["createrepo_c", build.path])

            modules_path = os.path.join(build.path, 'modules.yaml')
            repodata_path = os.path.join(build.path, 'repodata')
            mmd_index = Modulemd.ModuleIndex.new()
            mmd_index.add_module_stream(build.mmd)
            with open(modules_path, "w") as f:
                f.write(mmd_index.dump_to_string())
            check_call(['/usr/bin/modifyrepo_c', '--mdtype=modules', modules_path, repodata_path])
            success = True
        finally:
            if not success:
                shutil.rmtree(build.path)

    def _get_builds_recurse(self, builds, name, stream):
        if name == 'platform':
            # Pseudo-module doesn't quite exist
            return

        if name in builds:
            build = builds[name]
            if build.stream != stream:
                raise RuntimeError("Stream conflict for {}, both {} and {} are required",
                                   name, build.stream, stream)
            return build


        build = self.locate(name, stream)
        builds[name] = build

        dependencies = build.mmd.get_dependencies()
        # A built module should have its dependencies already expanded
        assert len(dependencies) <= 1

        if len(dependencies) == 1:
            for required_module in dependencies[0].get_runtime_modules():
                rs = dependencies[0].get_runtime_streams(required_module)
                # should already be expanded to a single stream
                assert len(rs) == 1
                self._get_builds_recurse(builds, required_module, rs[0])

    def get_builds(self, name, stream, version=None):
        builds = OrderedDict()

        self._get_builds_recurse(builds, name, stream)

        return builds



import click
import pipes
import re
import subprocess
import sys

def error(msg):
    click.secho('error: ', fg='red', bold=True, err=True, nl=False)
    click.echo(msg, err=True)

def die(msg):
    error(msg)
    sys.exit(1)

def warn(msg):
    click.secho('warning: ', fg='yellow', bold=True, err=True, nl=False)
    click.echo(msg, err=True)

def important(msg):
    click.secho(msg, err=True, bold=True)

def info(msg):
    click.secho('info: ', fg='blue', bold=True, err=True, nl=False)
    click.echo(msg, err=True)

def header(msg):
    important(msg)
    important('=' * len(msg))

def log_call(args):
    click.secho('running: ', fg='blue', bold=True, err=True, nl=False)
    click.echo(' '.join(pipes.quote(a) for a in args))

def check_call(args, cwd=None):
    log_call(args)
    rv = subprocess.call(args, cwd=cwd)
    if rv != 0:
        die("%s failed (exit status=%d)" % (args[0], rv))

class ModuleSpec(object):
    def __init__(self, name, stream, version=None, profile=None):
        self.name = name
        self.stream = stream
        self.version = version
        self.profile = profile

    def to_str(self, include_profile=True):
        result = self.name + ':' + self.stream
        if self.version:
            result += ':' + self.version
        if include_profile and self.profile:
            result += '/' + self.profile

        return result

    def __repr__(self):
        return "ModuleSpec({})".format(self.to_str())

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


def split_module_spec(module):
    # Current module naming guidelines are at:
    # https://docs.pagure.org/modularity/development/building-modules/naming-policy.html
    # We simplify the possible NAME:STREAM:CONTEXT:ARCH/PROFILE and only care about
    # NAME:STREAM or NAME:STREAM:VERSION with optional PROFILE. ARCH is determined by
    # the architecture. CONTEXT may become important in the future, but we ignore it
    # for now.
    #
    # Previously the separator was '-' instead of ':', which required hardcoding the
    # format of VERSION to distinguish between HYPHENATED-NAME-STREAM and NAME-STREAM-VERSION.
    # We support the old format for compatibility.
    #
    PATTERNS = [
        (r'^([^:/]+):([^:/]+):([^:/]+)(?:/([^:/]+))?$', 3, 4),
        (r'^([^:/]+):([^:/]+)(?:/([^:/]+))?$', None, 3),
        (r'^(.+)-([^-]+)-(\d{14})$', 3, None),
        (r'^(.+)-([^-]+)$', None, None)
    ]

    for pat, version_index, profile_index in PATTERNS:
        m = re.match(pat, module)
        if m:
            name = m.group(1)
            stream = m.group(2)
            version = None
            if version_index is not None:
                version = m.group(version_index)
            else:
                version = None
            if profile_index is not None:
                profile = m.group(profile_index)
            else:
                profile = None

            return ModuleSpec(name, stream, version, profile)

    raise RuntimeError(
        'Module specification should be NAME:STREAM[/PROFILE] or NAME:STREAM:VERSION[/PROFILE]. ' +
        '(NAME-STREAM and NAME-STREAM-VERSION supported for compatibility.)'
    )



import jinja2
import os
import yaml
import re
import shutil
import subprocess
from textwrap import dedent

from .flatpak_builder import FlatpakBuilder, FlatpakSourceInfo, FLATPAK_METADATA_ANNOTATIONS
from .module_locator import ModuleLocator
from .utils import check_call, die, log_call, warn, header, important, info, split_module_spec

class ContainerBuilder(object):
    def __init__(self, profile, containerspec, from_local=False, local_builds=[],
                 flatpak_metadata=FLATPAK_METADATA_ANNOTATIONS):
        self.profile = profile
        self.containerspec = os.path.abspath(containerspec)
        self.from_local = from_local
        self.local_builds = local_builds
        self.flatpak_metadata = flatpak_metadata

        with open(self.containerspec) as f:
            container_yaml = yaml.safe_load(f)

        compose_yaml = container_yaml.get('compose', {})
        modules = compose_yaml.get('modules', [])
        if not modules:
            die("No modules specified in the compose section of '{}'".format(self.containerspec))

        self.flatpak_yaml = container_yaml.get('flatpak', None)
        if not self.flatpak_yaml:
            die("No flatpak section in '{}'".format(containerspec))

        compose_yaml = container_yaml.get('compose', {})
        modules = compose_yaml.get('modules', [])
        if not modules:
            die("No modules specified in the compose section of '{}'".format(self.containerspec))

        if len(modules) > 1:
            warn("Multiple modules specified in compose section of '{}', using first".format(containerspec))

        self.module_spec = split_module_spec(modules[0])

    def _get_platform_version(self, builds):
        # Streams should already be expanded in the modulemd's that we retrieve
        #  modules were built against a particular dependency.
        def get_stream(module, req, req_streams):
            if len(req_streams) != 1:
                die("%s: stream list for '%s' is not expanded (%s)" %
                    (module.props.name, req, req_streams))
            return req_streams[0]

        platform_stream = None

        # Find the platform stream to get the base package set
        for build in builds.values():
            for dep in build.mmd.get_dependencies():
                for req in dep.get_runtime_modules():
                    if req == 'platform':
                        req_streams = dep.get_runtime_streams(req)
                        platform_stream = get_stream(build.mmd, req, req_streams)

        if platform_stream is None:
            die("Unable to determine base OS version from 'platform' module stream")

        m = re.match(self.profile.platform_stream_pattern, platform_stream)
        if m is None:
            die("'platform' module stream '{}' doesn't match '{}'".format(
                platform_stream, self.profile.platform_stream_pattern))

        return m.group(1)

    def build(self):
        header('BUILDING CONTAINER')
        important('container spec: {}'.format(self.containerspec))
        important('')

        module_build_id = self.module_spec.to_str(include_profile=False)

        locator = ModuleLocator(self.profile)
        if self.from_local:
            locator.add_local_build(module_build_id)
        for build_id in self.local_builds:
            locator.add_local_build(build_id)

        builds = locator.get_builds(self.module_spec.name, self.module_spec.stream, self.module_spec.version)
        base_build = list(builds.values())[0]

        builddir = os.path.expanduser("~/modulebuild/flatpaks")
        workdir = os.path.join(builddir, "{}-{}-{}".format(base_build.name, base_build.stream, base_build.version))
        if os.path.exists(workdir):
            info("Removing old output directory {}".format(workdir))
            shutil.rmtree(workdir)

        os.makedirs(workdir)
        info("Writing results to {}".format(workdir))

        has_modulemd = {}

        for build in builds.values():
            locator.ensure_downloaded(build)
            has_modulemd[build.name + ':' + build.stream] = build.has_module_metadata()

        repos = [build.yum_config() for build in builds.values()]

        source = FlatpakSourceInfo(self.flatpak_yaml, builds, base_build, self.module_spec.profile)

        builder = FlatpakBuilder(source, workdir, ".", flatpak_metadata=self.flatpak_metadata)

        component_label = self.flatpak_yaml.get('component', base_build.name)
        name_label = self.flatpak_yaml.get('name', base_build.name)
        version_label = base_build.stream
        release_label = base_build.version

        builder.add_labels({'name': name_label,
                            'com.redhat.component': component_label,
                            'version': version_label,
                            'release': release_label})

        env = jinja2.Environment(loader=jinja2.PackageLoader('flatpak_module_tools', 'templates'),
                                 autoescape=False)
        template = env.get_template('mock.cfg.j2')

        output_path = os.path.join(workdir, 'mock.cfg')

        # Check if DNF is new enough to support modules, if not, all packages from modular
        # repositories will automatically be enabled. If yes, we'll need to enable any
        # modular repositories - *if they actually have module metadata*
        try:
            subprocess.check_output(['dnf', 'module', 'list', '--enabled'],
                                    stderr=subprocess.STDOUT)
            have_dnf_module=True
        except subprocess.CalledProcessError:
            have_dnf_module=False

        platform_version = self._get_platform_version(builds)
        base_repo_url = self.profile.base_repo_url.format(platform=platform_version)
        template.stream(arch='x86_64',
                        base_repo_url=base_repo_url,
                        includepkgs=builder.get_includepkgs(),
                        repos=repos).dump(output_path)

        finalize_script = dedent("""\
            #!/bin/sh
            set -ex
            userdel -f mockbuild
            groupdel mock
            """ + builder.get_cleanup_script()) + dedent("""\
            cd /
            exec tar cf - --anchored --exclude='./sys/*' --exclude='./proc/*' --exclude='./dev/*' --exclude='./run/*' ./
        """)

        finalize_script_path = os.path.join(workdir, 'finalize.sh')
        with open(finalize_script_path, 'w') as f:
            f.write(finalize_script)
            os.fchmod(f.fileno(), 0o0755)

        info('Initializing installation path')
        check_call(['mock', '-q', '-r', output_path, '--clean'])

        # Using mock's config_opts['module_enable'] doesn't work because dnf tries
        # to enable modules before chroot_setup_cmd installs system-release, but
        # dnf needs /etc/os-release to figure out the platform module. So we do it this way.
        # https://github.com/rpm-software-management/mock/issues/232#issuecomment-456340663
        if have_dnf_module:
            info('Enabling modules')
            to_enable = [x for x in builder.get_enable_modules() if has_modulemd[x]]
            check_call(['mock', '-r', output_path, '--dnf-cmd', 'module', 'enable'] + to_enable)

        info('Installing packages')
        check_call(['mock', '-r', output_path, '--install'] + sorted(builder.get_install_packages()))

        info('Cleaning and exporting filesystem')
        check_call(['mock', '-q', '-r', output_path, '--copyin', finalize_script_path, '/root/finalize.sh'])

        builder.root = "."

        args = ['mock', '-q', '-r', output_path, '--shell', '/root/finalize.sh']
        log_call(args)
        process = subprocess.Popen(args, stdout=subprocess.PIPE)
        # When mock is using systemd-nspawn, systemd-nspawn dies with EPIPE if the output
        # stream is closed before it exits, even if the child of systemd-nspawn isn't
        # writing anything.
        # https://github.com/systemd/systemd/issues/11533
        filesystem_tar, manifestfile = builder._export_from_stream(process.stdout, close_stream=False)
        process.wait()
        process.stdout.close()
        if process.returncode != 0:
            die("finalize.sh failed (exit status=%d)" % process.returncode)

        ref_name, outfile, tarred_outfile = builder.build_container(filesystem_tar)

        local_outname = "{}-{}-{}.oci.tar.gz".format(base_build.name, base_build.stream, base_build.version)

        info('Compressing result')
        with open(local_outname, 'wb') as f:
            subprocess.check_call(['gzip', '-c', tarred_outfile], stdout=f)

        important('Created ' + local_outname)

        return local_outname



import koji
import json
import os
import re
import shutil
import subprocess
import tempfile

from six.moves.urllib.parse import urlparse

import click
import requests

from .utils import check_call, header, die

def _download_url(url, outdir):
    parts = urlparse(url)
    basename = os.path.basename(parts.path)
    if basename == '' or basename == 'oci':
        basename = "archive"

    path = os.path.join(outdir, basename)

    try:
        r = requests.get(url, stream=True)
        r.raise_for_status()
    except requests.exceptions.ConnectionError as e:
        die("Download failed for {}: {}".format(url, e))
    except requests.exceptions.HTTPError as e:
        die("Download failed: {}".format(e))

    total_length = int(r.headers.get('content-length'))

    with open(path, 'wb') as f:
        with click.progressbar(length=total_length, label="Downloading image") as pb:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    pb.update(len(chunk))
                    f.write(chunk)

    return path


def _download_koji_name_stream(profile, koji_name_stream, outdir):
    parts = koji_name_stream.split(':')
    if len(parts) == 1:
        name = parts[0]
        stream = None
    elif len(parts) == 2:
        name = parts[0]
        stream = parts[1]
    else:
        die("Koji download should be NAME or NAME:STREAM, not {}".format(koji_name_stream))

    options = koji.read_config(profile_name=profile.koji_profile, user_config=profile.koji_config)
    session_opts = koji.grab_session_options(options)
    session = koji.ClientSession(options['server'], session_opts)

    package_id = session.getPackageID(name)
    if package_id is None:
        die("Cannot find koji ID for {}".format(koji_name_stream))

    kwargs = {
        'type': 'image',
        'state': koji.BUILD_STATES['COMPLETE'],
        'queryOpts': { 'order': '-completion_ts'}
    }

    if stream is None:
        kwargs['queryOpts']['limit'] = 1

    builds = session.listBuilds(package_id, **kwargs)
    if stream is not None:
        builds = [b for b in builds if b['version'] == stream]

    if len(builds) == 0:
        die("Cannot find any Flatpak builds for {}".format(koji_name_stream))

    build = builds[0]
    archives = session.listArchives(build['build_id'])

    pathinfo = koji.PathInfo(topdir=options['topurl'])
    url = '/'.join((pathinfo.imagebuild(build), archives[0]['filename']))

    return _download_url(url, outdir)


class Installer(object):
    def __init__(self, profile):
        self.profile = profile
        self.source_koji_name_stream = None
        self.source_path = None
        self.source_url = None

        data_home = os.environ.get('XDG_DATA_HOME',
                                   os.path.expanduser('~/.local/share'))
        self.repodir = os.path.join(data_home, 'flatpak-module-tools', 'repo')

    def set_source_path(self, path):
        self.source_path = path

    def set_source_url(self, url):
        self.source_url = url

    def set_source_koji_name_stream(self, koji_name_stream):
        self.source_koji_name_stream = koji_name_stream

    def ensure_remote(self):
        if not os.path.exists(self.repodir):
            parent = os.path.dirname(self.repodir)
            if not os.path.exists(parent):
                os.makedirs(parent)

            check_call(['ostree', 'init', '--mode=archive-z2', '--repo', self.repodir])
            check_call(['flatpak', 'build-update-repo', self.repodir])

        output = subprocess.check_output(['flatpak', 'remotes', '--user'], encoding="UTF-8")
        if not re.search('^flatpak-module-tools\s', output, re.MULTILINE):
            check_call(['flatpak', 'remote-add',
                        '--user', '--no-gpg-verify',
                        'flatpak-module-tools', self.repodir])

    def install(self):
        header('INSTALLING')

        self.ensure_remote()

        try:
            workdir = tempfile.mkdtemp()
            ocidir = os.path.join(workdir, 'oci')
            os.mkdir(ocidir)

            if self.source_koji_name_stream:
                path = _download_koji_name_stream(self.profile, self.source_koji_name_stream, workdir)
            elif self.source_url:
                path = _download_url(self.source_url, workdir)
            elif self.source_path:
                path = os.path.abspath(self.source_path)
            else:
                raise RuntimeError("Installation source not set")

            check_call(['tar', 'xfa', path], cwd=ocidir)

            with open(os.path.join(ocidir, 'index.json')) as f:
                index_json = json.load(f)

            ref = index_json['manifests'][0]['annotations']['org.flatpak.ref']

            check_call(['flatpak', 'build-import-bundle',
                        '--update-appstream', '--oci',
                        self.repodir, ocidir])
        finally:
            shutil.rmtree(workdir)

        parts = ref.split('/')
        shortref = parts[0] + '/' + parts[1]

        try:
            with open(os.devnull, 'w') as devnull:
                old_origin = subprocess.check_output(['flatpak', 'info', '--user', '-o', shortref],
                                                     stderr=devnull, encoding="UTF-8").strip()
        except subprocess.CalledProcessError:
            old_origin = None

        if old_origin == 'flatpak-module-tools':
            check_call(['flatpak', 'update', '-y', '--user', ref])
        else:
            check_call(['flatpak', 'install', '-y', '--user', '--reinstall', 'flatpak-module-tools', ref])



import os

import pkg_resources
import yaml


_extra_config_files = []
def add_config_file(self, config_file):
    _extra_config_files.append(config_file)


_config = None
def get_config():
    global _config
    if _config is None:
        _config = Config()
        _config.read()

    return _config


_profile_name = None
def set_profile_name(profile_name):
    global _profile_name
    _profile_name = profile_name


def get_profile():
    return get_config().profiles[_profile_name]


class ProfileConfig(object):
    config_keys = [
        'base_repo_url',
        'koji_config',
        'koji_profile',
        'mbs_config_file',
        'mbs_config_section',
        'platform_stream_pattern',
    ]

    def __init__(self):
        for k in self.config_keys:
            setattr(self, k, None)

    def merge(self, yml):
        for k in self.config_keys:
            v = yml.get(k)
            if v is not None:
                setattr(self, k, v)

    def setdefaults(self, other):
        for k in self.config_keys:
            if getattr(self, k) is None:
                setattr(self, k, getattr(other, k))


class Config(object):
    def __init__(self):
        self.profiles = {}

    def _read_config_file(self, config_file):
        if isinstance(config_file, str):
            try:
                with open(config_file) as f:
                    yml = yaml.safe_load(f)
            except OSError:
                return
        else:
            yml = yaml.safe_load(config_file)

        for profile, profile_yml in yml['profiles'].items():
            if profile not in self.profiles:
                self.profiles[profile] = ProfileConfig()
            self.profiles[profile].merge(profile_yml)

    def read(self):
        config_files = []

        try:
            config_files += [
                pkg_resources.resource_stream('flatpak_module_tools', 'config.yaml')
            ]
        except:
            pass

        user_config = os.environ.get('XDG_CONFIG_HOME',
                                     os.path.expanduser('~/.config'))
        config_files += [
            '/etc/flatpak-module/config.yaml',
            '/etc/flatpak-module/config.d/',
            os.path.join(user_config, 'flatpak-module/config.yaml'),
            os.path.join(user_config, 'flatpak-module/config.d/'),
        ]

        config_files += _extra_config_files

        for config_file in config_files:
            if isinstance(config_file, str) and config_file.endswith('/'):
                try:
                    files = os.listdir(config_file)
                except OSError:
                    continue
                for f in sorted(files, reverse=True):
                    if f.endswith('yaml'):
                        self._read_config_file(os.path.join(config_file, f))
            else:
                self._read_config_file(config_file)

        defaults = self.profiles.get('__default__')
        if defaults is not None:
            for t in self.profiles.values():
                if t is not defaults:
                    t.setdefaults(defaults)



"""
Copyright (c) 2017 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.

This file is code for generating Flatpak OCI images out of a filesystem
image. It is shared between:

 https://github.com/projectatomic/atomic-reactor
 https://pagure.io/flatpak-module-tools
"""

import errno
import hashlib
import json
import logging
import os
from six.moves import configparser
import re
import shlex
import shutil
import subprocess
import sys
import tarfile
from textwrap import dedent
from xml.etree import ElementTree


FLATPAK_METADATA_LABELS = "labels"
FLATPAK_METADATA_ANNOTATIONS = "annotations"
FLATPAK_METADATA_BOTH = "both"


# Work around lack of RawConfigParser space_around_delimeters in Python-2.7
if sys.version_info[0] == 2:
    class MyConfigParser(configparser.RawConfigParser):
        # Copy from configparser.RawConfigParser - Copyright (c) 2017 Python Software Foundation
        # https://www.python.org/download/releases/2.7/license/
        def write(self, fp, space_around_delimiters=False):
            """Write an .ini-format representation of the configuration state."""
            for section in self._sections:
                fp.write("[%s]\n" % section)
                for (key, value) in self._sections[section].items():
                    if key == "__name__":
                        continue
                    if (value is not None) or (self._optcre == self.OPTCRE):
                        key = "=".join((key, str(value).replace('\n', '\n\t')))
                    fp.write("%s\n" % (key))
                fp.write("\n")

    CP = MyConfigParser
else:
    CP = configparser.RawConfigParser


# Returns flatpak's name for the current arch
def get_arch():
    return subprocess.check_output(['flatpak', '--default-arch'],
                                   universal_newlines=True).strip()


# flatpak build-init requires the sdk and runtime to be installed on the
# build system (so that subsequent build steps can execute things with
# the SDK). While it isn't impossible to download the runtime image and
# install the flatpak, that would be a lot of unnecessary complexity
# since our build step is just unpacking the filesystem we've already
# created. This is a stub implementation of 'flatpak build-init' that
# doesn't check for the SDK or use it to set up the build filesystem.
def build_init(directory, appname, sdk, runtime, runtime_branch, tags=[]):
    if not os.path.isdir(directory):
        os.mkdir(directory)
    with open(os.path.join(directory, "metadata"), "w") as f:
        f.write(dedent("""\
                       [Application]
                       name={appname}
                       runtime={runtime}/{arch}/{runtime_branch}
                       sdk={sdk}/{arch}/{runtime_branch}
                       """.format(appname=appname,
                                  sdk=sdk,
                                  runtime=runtime,
                                  runtime_branch=runtime_branch,
                                  arch=get_arch())))
        if tags:
            f.write("tags=" + ";".join(tags) + "\n")
    os.mkdir(os.path.join(directory, "files"))


class ModuleInfo(object):
    def __init__(self, name, stream, version, mmd, rpms):
        self.name = name
        self.stream = stream
        self.version = version
        self.mmd = mmd
        self.rpms = rpms


class FileTreeProcessor(object):
    def __init__(self, builddir, flatpak_yaml):
        self.app_root = os.path.join(builddir, "files")
        self.app_id = flatpak_yaml['id']

        self.appdata_license = flatpak_yaml.get('appdata-license', None)
        self.appstream_compose = flatpak_yaml.get('appstream-compose', True)
        self.copy_icon = flatpak_yaml.get('copy-icon', False)
        self.desktop_file_name_prefix = flatpak_yaml.get('desktop-file-name-prefix')
        self.desktop_file_name_suffix = flatpak_yaml.get('desktop-file-name-suffix')
        self.rename_appdata_file = flatpak_yaml.get('rename-appdata-file', None)
        self.rename_desktop_file = flatpak_yaml.get('rename-desktop-file')
        self.rename_icon = flatpak_yaml.get('rename-icon')

        self.log = logging.getLogger(__name__)

    def _find_appdata_file(self):
        # We order these so that share/appdata/XXX.appdata.xml if found
        # first, as this is the target name, and apps may have both, which will
        # cause issues with the rename.
        extensions = [
            ".appdata.xml",
            ".metainfo.xml",
        ]

        dirs = [
            "share/appdata",
            "share/metainfo",
        ]

        for d in dirs:
            appdata_dir = os.path.join(self.app_root, d)
            for ext in extensions:
                if self.rename_appdata_file is not None:
                    basename = self.rename_appdata_file
                else:
                    basename = self.app_id + ext

                source = os.path.join(appdata_dir, basename)
                if os.path.exists(source):
                    return source

        return None

    def _rewrite_appdata(self):
        tree = ElementTree.parse(self.appdata_file)

        # replace component/id
        n_root = tree.getroot()
        if n_root.tag != "component" and n_root.tag != "application":
            raise RuntimeError("Root node is not <application> or <component>")

        n_license = n_root.find("project_license")
        if n_license is not None:
            n_license.text = self.appdata_license

        tree.write(self.appdata_file, encoding="UTF-8", xml_declaration=True)

    def _process_appdata_file(self):
        appdata_source = self._find_appdata_file ()
        self.appdata_file = None

        if appdata_source:
            # We always use the old name / dir, in case the runtime has older appdata tools
            appdata_dir = os.path.join(self.app_root, "share", "appdata")
            appdata_basename = self.app_id + ".appdata.xml"
            self.appdata_file = os.path.join(appdata_dir, appdata_basename)

            if appdata_source != self.appdata_file:
                src_basename = os.path.basename(appdata_source)
                self.log.info("Renaming %s to share/appdata/%s", src_basename, appdata_basename)

                if not os.path.exists(appdata_dir):
                    os.makedirs(appdata_dir)
                os.rename(appdata_source, self.appdata_file)

            if self.appdata_license:
                self._rewrite_appdata()

    def _rename_desktop_file(self):
        if not self.rename_desktop_file:
            return

        applications_dir = os.path.join(self.app_root, "share", "applications")
        src = os.path.join(applications_dir, self.rename_desktop_file)
        desktop_basename = self.app_id + ".desktop"
        dest = os.path.join(applications_dir, desktop_basename)

        self.log.info("Renaming %s to %s", self.rename_desktop_file, desktop_basename)
        os.rename(src, dest)

        if self.appdata_file:
            tree = ElementTree.parse(self.appdata_file)

            # replace component/id
            n_root = tree.getroot()
            if n_root.tag != "component" and n_root.tag != "application":
                raise RuntimeError("Root node is not <application> or <component>")

            n_id = n_root.find("id")
            if n_id is not None:
                if n_id.text == self.rename_desktop_file:
                    n_id.text = self.app_id

            # replace any optional launchable
            n_launchable = n_root.find("launchable")
            if n_launchable is not None:
                if n_launchable.text == self.rename_desktop_file:
                    n_launchable.text = desktop_basename

            tree.write(self.appdata_file, encoding="UTF-8", xml_declaration=True)

    def _rename_icon(self):
        if not self.rename_icon:
            return

        found_icon = False
        icons_dir = os.path.join(self.app_root, "share/icons")

        for full_dir, dirnames, filenames in os.walk(icons_dir):
            relative = full_dir[len(icons_dir):]
            depth = relative.count("/")

            for source_file in filenames:
                if source_file.startswith(self.rename_icon):
                    source_path = os.path.join(full_dir, source_file)
                    is_file = os.path.isfile(source_path)
                    extension = source_file[len(self.rename_icon):]

                    if is_file and depth == 3 and (extension.startswith(".") or
                                                   extension.startswith("-symbolic")):
                        found_icon = True
                        new_name = self.app_id + extension

                        self.log.info("%s icon %s/%s to %s/%s",
                                      "Copying" if self.copy_icon else "Renaming",
                                      relative[1:], source_file,
                                      relative[1:], new_name)

                        dest_path = os.path.join(full_dir, new_name)
                        if self.copy_icon:
                            shutil.copy(source_path, dest_path)
                        else:
                            os.rename(source_path, dest_path)
                    else:
                        if not is_file:
                            self.log.debug("%s/%s matches 'rename-icon', but not a regular file",
                                           full_dir, source_file)
                        elif depth != 3:
                            self.log.debug("%s/%s matches 'rename-icon', but not at depth 3",
                                           full_dir, source_file)
                        else:
                            self.log.debug("%s/%s matches 'rename-icon', but name does not continue with '.' or '-symbolic.'",
                                           full_dir, source_file)

        if not found_icon:
            raise RuntimeError("icon {} not found below {}".format(self.rename_icon, icons_dir))

    def _rewrite_desktop_file(self):
        if not (self.rename_icon or self.desktop_file_name_prefix or self.desktop_file_name_suffix):
            return

        applications_dir = os.path.join(self.app_root, "share/applications")
        desktop_basename = self.app_id + ".desktop"
        desktop = os.path.join(applications_dir, desktop_basename)

        self.log.debug("Rewriting contents of %s", desktop_basename)

        cp = CP()
        cp.optionxform = str
        with open(desktop) as f:
            cp.readfp(f)

        desktop_keys = cp.options('Desktop Entry')

        if self.rename_icon:
            if self.rename_icon:
                original_icon_name = cp.get('Desktop Entry', 'Icon')
                cp.set('Desktop Entry', 'Icon', self.app_id)

                for key in desktop_keys:
                    if key.startswith("Icon["):
                        if cp.get('Desktop Entry', key) == original_icon_name:
                            cp.set('Desktop Entry', key, self.app_id)

        if self.desktop_file_name_suffix or self.desktop_file_name_prefix:
            for key in desktop_keys:
                if key == "Name" or key.startswith("Name["):
                    name = cp.get('Desktop Entry', key)
                    new_name = ((self.desktop_file_name_prefix or "") +
                                name +
                                (self.desktop_file_name_suffix or ""))
                    cp.set('Desktop Entry', key, new_name)

        with open(desktop, "w") as f:
            cp.write(f, space_around_delimiters=False)

    def _compose_appstream(self):
        if not self.appstream_compose or not self.appdata_file:
            return

        subprocess.check_call(['appstream-compose',
                               '--verbose',
                               '--prefix', self.app_root,
                               '--basename', self.app_id,
                               '--origin', 'flatpak',
                               self.app_id])

    def process(self):
        self._process_appdata_file()
        self._rename_desktop_file()
        self._rename_icon()
        self._rewrite_desktop_file()
        self._compose_appstream()

class FlatpakSourceInfo(object):
    def __init__(self, flatpak_yaml, modules, base_module, profile=None):
        self.flatpak_yaml = flatpak_yaml
        self.modules = modules
        self.base_module = base_module

        # A runtime module must have a 'runtime' profile, but can have other
        # profiles for SDKs, minimal runtimes, etc.
        self.runtime = 'runtime' in base_module.mmd.get_profile_names()

        if profile:
            self.profile = profile
        elif self.runtime:
            self.profile = 'runtime'
        else:
            self.profile = 'default'

        assert self.profile in base_module.mmd.get_profile_names()

    # The module for the Flatpak runtime that this app runs against
    @property
    def runtime_module(self):
        assert not self.runtime

        dependencies = self.base_module.mmd.get_dependencies()
        # A built module should have its dependencies already expanded
        assert len(dependencies) <= 1

        if len(dependencies) == 1:
            for module_name in dependencies[0].get_buildtime_modules():
                try:
                    module = self.modules[module_name]
                    if 'runtime' in module.mmd.get_profile_names():
                        return module
                except KeyError:
                    pass

        raise RuntimeError("Failed to identify runtime module in the buildrequires for {}"
                           .format(self.base_module.name))

    # All modules that were build against the Flatpak runtime,
    # and thus were built with prefix=/app. This is primarily the app module
    # but might contain modules shared between multiple flatpaks as well.
    @property
    def app_modules(self):
        runtime_module_name = self.runtime_module.mmd.props.module_name

        def is_app_module(m):
            dependencies = m.mmd.get_dependencies()
            return runtime_module_name in dependencies[0].get_buildtime_modules()

        return [m for m in self.modules.values() if is_app_module(m)]


class FlatpakBuilder(object):
    def __init__(self, source, workdir, root, parse_manifest=None,
                 flatpak_metadata=FLATPAK_METADATA_ANNOTATIONS):
        self.source = source
        self.workdir = workdir
        self.root = root
        self.parse_manifest = parse_manifest

        if not flatpak_metadata in (FLATPAK_METADATA_ANNOTATIONS,
                                    FLATPAK_METADATA_LABELS,
                                    FLATPAK_METADATA_BOTH):
            raise ValueError("Bad flatpak_metadata value %s" % flatpak_metadata)
        self.flatpak_metadata = flatpak_metadata

        self.log = logging.getLogger(__name__)

        self.extra_labels = {}

    def add_labels(self, labels):
        """Specify additional labels to add to the generated image"""

        self.log.warning(labels)
        self.extra_labels.update(labels)

    def precheck(self):
        source = self.source

        # For a runtime, certain information is duplicated between the container.yaml
        # and the modulemd, check that it matches
        if source.runtime:
            flatpak_yaml = source.flatpak_yaml
            flatpak_xmd = self.source.base_module.mmd.get_xmd()['flatpak']

            def check(condition, what):
                if not condition:
                    raise RuntimeError(
                        "Mismatch for {} betweeen module xmd and container.yaml".format(what))

            check(flatpak_yaml['branch'] == flatpak_xmd['branch'], "'branch'")
            check(source.profile in flatpak_xmd['runtimes'], 'profile name')

            profile_xmd = flatpak_xmd['runtimes'][source.profile]

            check(flatpak_yaml['id'] == profile_xmd['id'], "'id'")
            check(flatpak_yaml.get('runtime', None) ==
                  profile_xmd.get('runtime', None), "'runtime'")
            check(flatpak_yaml.get('sdk', None) == profile_xmd.get('sdk', None), "'sdk'")

    def get_enable_modules(self):
        # We need to enable all the modules other than the platform pseudo-module
        # sorted for testability.
        return sorted(m.mmd.props.module_name + ':' + m.mmd.props.stream_name
                      for m in self.source.modules.values()
                      if m.mmd.props.module_name != 'platform')

    def get_install_packages(self):
        source = self.source
        packages = source.base_module.mmd.get_profile(source.profile).get_rpms()
        if not source.runtime:
            # The flatpak-runtime-config package is needed when building an application
            # Flatpak because it includes file triggers for files in /app. (Including just
            # this package avoids having to install the entire runtime package set; if
            # we need to make this configurable it could be a separate profile of
            # the runtime.)
            packages.append('flatpak-runtime-config')

        return packages

    def get_includepkgs(self):
        source = self.source

        # For a runtime, we want to make sure that the set of RPMs that is installed
        # into the filesystem is *exactly* the set that is listed in the runtime
        # profile. Requiring the full listed set of RPMs to be listed makes it
        # easier to catch unintentional changes in the package list that might break
        # applications depending on the runtime. It also simplifies the checking we
        # do for application flatpaks, since we can simply look at the runtime
        # modulemd to find out what packages are present in the runtime.
        #
        # For an application, we want to make sure that each RPM that is installed
        # into the filesystem is *either* an RPM that is part of the 'runtime'
        # profile of the base runtime, or from a module that was built with
        # flatpak-rpm-macros in the install root and, thus, prefix=/app.
        #
        # We achieve this by restricting the set of available packages in the dnf
        # configuration to just the ones that we want.
        #
        # The advantage of doing this upfront, rather than just checking after the
        # fact is that this makes sure that when a application is being installed,
        # we don't get a different package to satisfy a dependency than the one
        # in the runtime - e.g. aajohan-comfortaa-fonts to satisfy font(:lang=en)
        # because it's alphabetically first.

        if not source.runtime:
            runtime_module = source.runtime_module
            runtime_profile = runtime_module.mmd.get_profile('runtime')
            available_packages = sorted(runtime_profile.get_rpms())

            for m in source.app_modules:
                # Strip off the '.rpm' suffix from the filename to get something
                # that DNF can parse.
                available_packages.extend(x[:-4] for x in m.rpms)
        else:
            base_module = source.base_module
            available_packages = sorted(base_module.mmd.get_profile(source.profile).get_rpms())

        return available_packages

    def get_cleanup_script(self):
        cleanup_commands = self.source.flatpak_yaml.get('cleanup-commands')
        if cleanup_commands is not None:
            return cleanup_commands.rstrip() + '\n'
        else:
            return ''

    # Compiles a list of path mapping rules to a simple function that matches
    # against a list of fixed patterns, see below for rule syntax
    def _compile_target_rules(self, rules):
        patterns = []
        for source, target in rules:
            source = re.sub("^ROOT", self.root, source)
            if source.endswith("/"):
                patterns.append((re.compile(source + "(.*)"), target, False))
                patterns.append((source[:-1], target, True))
            else:
                patterns.append((source, target, True))

        def get_target_func(path):
            for source, target, is_exact_match in patterns:
                if is_exact_match:
                    if source == path:
                        return target
                else:
                    m = source.match(path)
                    if m:
                        return os.path.join(target, m.group(1))

            return None

        return get_target_func

    # Rules for mapping paths within the exported filesystem image to their
    # location in the final flatpak filesystem
    #
    # ROOT = root of the filesystem to extract - e.g. var/tmp/flatpak-build
    # No trailing slash - map a directory itself exactly
    # trailing slash - map a directory and everything inside of it

    def _get_target_path_runtime(self):
        return self._compile_target_rules([
            # We need to make sure that 'files' is created before 'files/etc',
            # which wouldn't happen if just relied on ROOT/usr/ => files.
            # Instead map ROOT => files and omit ROOT/usr
            ("ROOT", "files"),
            ("ROOT/usr", None),

            # We map ROOT/usr => files and ROOT/etc => files/etc. This creates
            # A conflict between ROOT/usr/etc and /ROOT/etc. Just assume there
            # is nothing useful in /ROOT/usr/etc.
            ("ROOT/usr/etc/", None),

            ("ROOT/usr/", "files"),
            ("ROOT/etc/", "files/etc")
    ])

    def _get_target_path_app(self):
        return self._compile_target_rules([
            ("ROOT/app/", "files")
        ])

    def _export_from_stream(self, export_stream, close_stream=True):
        if self.source.runtime:
            get_target_path = self._get_target_path_runtime()
        else:
            get_target_path = self._get_target_path_app()

        outfile = os.path.join(self.workdir, 'filesystem.tar.gz')
        manifestfile = os.path.join(self.workdir, 'flatpak-build.rpm_qf')

        out_fileobj = open(outfile, "wb")
        compress_process = subprocess.Popen(['gzip', '-c'],
                                            stdin=subprocess.PIPE,
                                            stdout=out_fileobj)
        in_tf = tarfile.open(fileobj=export_stream, mode='r|')
        out_tf = tarfile.open(fileobj=compress_process.stdin, mode='w|')

        for member in in_tf:
            if member.name == 'var/tmp/flatpak-build.rpm_qf':
                reader = in_tf.extractfile(member)
                with open(manifestfile, 'wb') as out:
                    out.write(reader.read())
                reader.close()
            target_name = get_target_path(member.name)
            if target_name is None:
                continue

            # Match the ownership/permissions changes done by 'flatpak build-export'.
            # See commit_filter() in:
            #   https://github.com/flatpak/flatpak/blob/master/app/flatpak-builtins-build-export.c
            #
            # We'll run build-export anyways in the app case, but in the runtime case we skip
            # flatpak build-export and use ostree directly.
            member.uid = 0
            member.gid = 0
            member.uname = "root"
            member.gname = "root"

            if member.isdir():
                member.mode = 0o0755
            elif member.isreg():
                if member.mode & 0o0100:
                    member.mode = 0o0755
                else:
                    member.mode = 0o0644

            member.name = target_name
            if member.islnk():
                # Hard links have full paths within the archive (no leading /)
                link_target = get_target_path(member.linkname)
                if link_target is None:
                    self.log.debug("Skipping %s, hard link to %s", target_name, link_target)
                    continue
                member.linkname = link_target
                out_tf.addfile(member)
            elif member.issym():
                # Symlinks have the literal link target, which will be
                # relative to the chroot and doesn't need rewriting
                out_tf.addfile(member)
            else:
                f = in_tf.extractfile(member)
                out_tf.addfile(member, fileobj=f)

        in_tf.close()
        out_tf.close()
        if close_stream:
            export_stream.close()
        compress_process.stdin.close()
        if compress_process.wait() != 0:
            raise RuntimeError("gzip failed")
        out_fileobj.close()

        return outfile, manifestfile

    def _get_components(self, manifest):
        with open(manifest, 'r') as f:
            lines = f.readlines()

        return self.parse_manifest(lines)

    def _filter_app_manifest(self, components):
        runtime_rpms = set(self.source.runtime_module.mmd.get_profile('runtime').get_rpms())

        return [c for c in components if not c['name'] in runtime_rpms]

    def get_components(self, manifest):
        all_components = self._get_components(manifest)
        if self.source.runtime:
            image_components = all_components
        else:
            image_components = self._filter_app_manifest(all_components)

        return image_components

    def _build_finish(self, builddir):
        info = self.source.flatpak_yaml

        finish_args = []
        if 'finish-args' in info:
            # shlex.split(None) reads from standard input, so avoid that
            finish_args = shlex.split(info['finish-args'] or '', comments=True)
        if 'command' in info and not self.source.runtime:
            finish_args = ['--command', info['command']] + finish_args

        subprocess.check_call(['flatpak', 'build-finish'] + finish_args + [builddir])

    def _create_runtime_oci(self, tarred_filesystem, outfile):
        info = self.source.flatpak_yaml

        builddir = os.path.join(self.workdir, "build")
        os.mkdir(builddir)

        filesdir = os.path.join(builddir, "files")
        os.mkdir(filesdir)

        repo = os.path.join(self.workdir, "repo")
        subprocess.check_call(['ostree', 'init', '--mode=archive-z2', '--repo', repo])

        id_ = info['id']
        runtime_id = info.get('runtime', id_)
        sdk_id = info.get('sdk', id_)
        branch = info['branch']

        args = {
            'id': id_,
            'runtime_id': runtime_id,
            'sdk_id': sdk_id,
            'arch': get_arch(),
            'branch': branch
        }

        METADATA_TEMPLATE = dedent("""\
            [Runtime]
            name={id}
            runtime={runtime_id}/{arch}/{branch}
            sdk={sdk_id}/{arch}/{branch}
            """)

        with open(os.path.join(builddir, 'metadata'), 'w') as f:
            f.write(METADATA_TEMPLATE.format(**args))

        # Run flatpak-build-finish to add extra metadata, based on finish-args
        self._build_finish(builddir)

        runtime_ref = 'runtime/{id}/{arch}/{branch}'.format(**args)

        subprocess.check_call(['ostree', 'commit',
                               '--repo', repo, '--owner-uid=0',
                               '--owner-gid=0', '--no-xattrs',
                               '--canonical-permissions',
                               '--branch', runtime_ref,
                               '-s', 'build of ' + runtime_ref,
                               '--tree=tar=' + tarred_filesystem,
                               '--tree=dir=' + builddir])
        subprocess.check_call(['ostree', 'summary', '-u', '--repo', repo])

        subprocess.check_call(['flatpak', 'build-bundle', repo,
                               '--oci', '--runtime',
                               outfile, id_, branch])

        return runtime_ref

    def _find_runtime_info(self):
        runtime_module = self.source.runtime_module

        flatpak_xmd = runtime_module.mmd.get_xmd()['flatpak']
        runtime_id = flatpak_xmd['runtimes']['runtime']['id']
        sdk_id = flatpak_xmd['runtimes']['runtime'].get('sdk', runtime_id)
        runtime_version = flatpak_xmd['branch']

        return runtime_id, sdk_id, runtime_version

    def _create_app_oci(self, tarred_filesystem, outfile):
        info = self.source.flatpak_yaml
        app_id = info['id']
        app_branch = info.get('branch', 'master')

        builddir = os.path.join(self.workdir, "build")
        os.mkdir(builddir)

        repo = os.path.join(self.workdir, "repo")

        runtime_id, sdk_id, runtime_version = self._find_runtime_info()

        # See comment for build_init() for why we can't use 'flatpak build-init'
        # subprocess.check_call(['flatpak', 'build-init',
        #                        builddir, app_id, runtime_id, runtime_id, runtime_version])
        build_init(builddir, app_id, sdk_id, runtime_id, runtime_version, tags=info.get('tags', []))

        # with gzip'ed tarball, tar is several seconds faster than tarfile.extractall
        subprocess.check_call(['tar', 'xCfz', builddir, tarred_filesystem])

        processor = FileTreeProcessor(builddir, info)
        processor.process()

        self._build_finish(builddir)

        # If we don't have a working bubblewrap, then we need to pass --disable-sandbox
        # to 'flatpak build-export'. On some systems, bwrap may be part
        # of the Flatpak install.

        def try_bwrap(exec_name):
            # return: True - works; False - doesn't work; None - not found
            try:
                return subprocess.call([exec_name, '--bind', '/', '/', 'true'],
                                       stdout=devnull, stderr=devnull) == 0
            except OSError as e:
                if e.errno == errno.ENOENT:
                    return None
                raise

        def try_export(disable_sandbox):
            args = ['flatpak', 'build-export', repo, builddir, app_branch]
            if disable_sandbox:
                args += ['--disable-sandbox']
            subprocess.check_call(args)

        with open(os.devnull) as devnull:
            have_bwrap = try_bwrap('bwrap')
            if have_bwrap is None:
                have_bwrap = try_bwrap('/usr/libexec/flatpak-bwrap')

        if have_bwrap:
            try_export(disable_sandbox=False)
        else:
            self.log.info('No working bubblewrap, callling flatpak build-export --disable-sandbox')
            try:
                try_export(disable_sandbox=True)
            except subprocess.CalledProcessError:
                # Older flatpak without --disable-sandbox?
                self.log.info('Retrying without --disable-sandbox')
                try_export(disable_sandbox=False)

        subprocess.check_call(['flatpak', 'build-bundle', repo, '--oci',
                               outfile, app_id, app_branch])

        app_ref = 'app/{app_id}/{arch}/{branch}'.format(app_id=app_id,
                                                        arch=get_arch(),
                                                        branch=app_branch)

        return app_ref

    # Update the config JSON for the image:
    #
    #  * Convert OCI images with labels (Flatpak >= 1.6) to OCI images
    #    with annotations (Flaptak < 1.6), and vice-versa, or generate
    #    compat images with both labels and annotations.
    #  * Add in extra labels specified by the caller
    #  * Add a history entry if missing - old versions of Flatpak write an
    #    image config file without any history entries.
    #    History isn't required by OCI spec, but multiple tools have problems
    #    without history entries matching the layers of the image.
    #    https://github.com/flatpak/flatpak/commit/be9961ecf65a081aac24e2007a0af7be1424eb38
    def _fixup_config(self, outfile):
        def get_path_from_descriptor(descriptor):
            assert descriptor["digest"].startswith("sha256:")
            return os.path.join(outfile, "blobs", "sha256", descriptor["digest"][len("sha256:"):])

        def update_descriptor(descriptor, contents_json):
            old_path = get_path_from_descriptor(descriptor)
            contents_bytes = json.dumps(contents_json, indent=4).encode("UTF-8")
            digest = hashlib.sha256(contents_bytes).hexdigest()
            descriptor["digest"] = "sha256:" + digest
            descriptor["size"] = len(contents_bytes)
            new_path = get_path_from_descriptor(descriptor)
            with open(new_path, "wb") as f:
                f.write(contents_bytes)
            os.remove(old_path)

        with open(os.path.join(outfile, "index.json")) as f:
            index_json = json.load(f)

        old_manifest = get_path_from_descriptor(index_json["manifests"][0])
        with open(old_manifest) as f:
            manifest_json = json.load(f)
        annotations = manifest_json.setdefault("annotations", {})
        old_config = get_path_from_descriptor(manifest_json["config"])
        with open(old_config) as f:
            config_json = json.load(f)

        if config_json.get("history") is None:
            self.log.warning("No history in the image config, adding one")
            config_json["history"] = [
                {
                    "created": config_json["created"],
                    "created_by": "flatpak build-bundle",
                },
            ]

        config = config_json.setdefault("config", {})
        labels = config.setdefault("Labels", {})

        if self.flatpak_metadata != FLATPAK_METADATA_ANNOTATIONS:
            # Merge in the annotations as labels
            to_delete = list()
            for k, v in annotations.items():
                if k.startswith("org.flatpak.") or k.startswith("org.freedesktop."):
                    if not k in labels:
                        labels[k] = v
                    if self.flatpak_metadata != FLATPAK_METADATA_BOTH:
                        to_delete.append(k)

            for k in to_delete:
                del annotations[k]

        if self.flatpak_metadata != FLATPAK_METADATA_LABELS:
            # And merge the labels as annotations the other way
            to_delete = list()
            for k, v in labels.items():
                if k.startswith("org.flatpak.") or k.startswith("org.freedesktop."):
                    if not k in annotations:
                        annotations[k] = v
                    if self.flatpak_metadata != FLATPAK_METADATA_BOTH:
                        to_delete.append(k)

            for k in to_delete:
                del labels[k]

        labels.update(self.extra_labels)

        update_descriptor(manifest_json["config"], config_json)
        update_descriptor(index_json["manifests"][0], manifest_json)

        with open(os.path.join(outfile, "index.json"), "w") as f:
            json.dump(index_json, f, indent=4)

    def build_container(self, tarred_filesystem):
        outfile = os.path.join(self.workdir, 'flatpak-oci-image')

        if self.source.runtime:
            ref_name = self._create_runtime_oci(tarred_filesystem, outfile)
        else:
            ref_name = self._create_app_oci(tarred_filesystem, outfile)

        self._fixup_config(outfile)

        tarred_outfile = outfile + '.tar'
        with tarfile.TarFile(tarred_outfile, "w") as tf:
            for f in os.listdir(outfile):
                tf.add(os.path.join(outfile, f), f)

        return ref_name, outfile, tarred_outfile



import koji

options = koji.read_config(profile_name='koji')
session_opts = koji.grab_session_options(options)
session = koji.ClientSession(options['server'], session_opts)

session.CGImport({'a': 'Hydrogen est une boîte à rythme.'}, 'foo')



import logging

import click

from .config import add_config_file, set_profile_name, get_profile
from .container_builder import ContainerBuilder
from .flatpak_builder import (FLATPAK_METADATA_ANNOTATIONS,
                              FLATPAK_METADATA_BOTH,
                              FLATPAK_METADATA_LABELS)
from .installer import Installer
from .module_builder import ModuleBuilder
from .utils import die

@click.group()
@click.option('-v', '--verbose', is_flag=True,
              help='Show verbose debugging output')
@click.option('-c', '--config', metavar='CONFIG_YAML', multiple=True,
              help='Additional configuration file to read')
@click.option('-p', '--profile', metavar='PROFILE_NAME', default='production',
              help='Alternate configuration profile to use')
def cli(verbose, config, profile):
    for c in reversed(config):
        add_config_file(c)

    set_profile_name(profile)
    try:
        get_profile()
    except KeyError:
        die("Unknown profile '{}'".format(profile))

    if verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)


@cli.command(name="local-build")
@click.option('--add-local-build', metavar='BUILD_ID', multiple=True,
              help='include a local MBS module build  as a source for the build')
@click.option('--containerspec', metavar='CONTAINER_YAML', default='./container.yaml',
              help='path to container.yaml - defaults to ./container.yaml')
@click.option('--modulemd', metavar='MODULEMD',
              help='path to modulemd file')
@click.option('--stream', metavar='STREAM',
              help='module stream for the build')
@click.option('--install', is_flag=True,
              help='automatically install Flatpak for the current user')
def local_build(add_local_build, containerspec, modulemd, stream, install):
    """Build module locally, then build a container"""

    module_builder = ModuleBuilder(profile=get_profile(),
                                   modulemd=modulemd, stream=stream,
                                   local_builds=add_local_build)
    container_builder = ContainerBuilder(profile=get_profile(),
                                         containerspec=containerspec,
                                         local_builds=add_local_build,
                                         from_local=True)

    if (container_builder.module_spec.name != module_builder.name or
        container_builder.module_spec.stream != module_builder.stream):
        die("Module will be built as {}:{}, but container.yaml calls for {}"
            .format(module_builder.name, module_builder.stream,
                    container_builder.module_spec.to_str(include_profile=False)))

    module_builder.build()

    tarfile = container_builder.build()

    if install:
        installer = Installer(profile=get_profile())
        installer.set_source_path(tarfile)
        installer.install()


@cli.command(name="build-module")
@click.option('--add-local-build', metavar='BUILD_ID', multiple=True,
              help='include a local MBS module build  as a source for the build')
@click.option('--modulemd', metavar='MODULEMD',
              help='Path to modulemd file')
@click.option('--stream', metavar='STREAM',
              help='module stream for the build')
def build_module(add_local_build, modulemd, stream):
    """Build module locally"""

    module_builder = ModuleBuilder(profile=get_profile(),
                                   modulemd=modulemd, stream=stream,
                                   local_builds=add_local_build)
    module_builder.build()


@cli.command(name="build-container")
@click.option('--add-local-build', metavar='BUILD_ID', multiple=True,
              help='include a local MBS module build  as a source for the build')
@click.option('--from-local', is_flag=True,
              help='Use a local build for the module source listed in container.yaml ')
@click.option('--flatpak-metadata',
              type=click.Choice([FLATPAK_METADATA_LABELS,
                                 FLATPAK_METADATA_ANNOTATIONS,
                                 FLATPAK_METADATA_BOTH], case_sensitive=False),
              default=FLATPAK_METADATA_ANNOTATIONS,
              help='How to store Flatpak metadata in the container')
@click.option('--containerspec', metavar='CONTAINER_YAML', default='./container.yaml',
              help='Path to container.yaml - defaults to ./container.yaml')
@click.option('--install', is_flag=True,
              help='automatically install Flatpak for the current user')
def build_container(add_local_build, from_local, flatpak_metadata, containerspec, install):
    """Build a container from local or remote module"""

    container_builder = ContainerBuilder(profile=get_profile(),
                                         containerspec=containerspec,
                                         local_builds=add_local_build,
                                         from_local=from_local,
                                         flatpak_metadata=flatpak_metadata)
    tarfile = container_builder.build()

    if install:
        installer = Installer(profile=get_profile())
        installer.set_source_path(tarfile)
        installer.install()


@cli.command()
@click.option('--koji', is_flag=True,
              help='Look up argument as NAME[:STREAM] in Koji')
@click.argument('path_or_url')
def install(koji, path_or_url):
    """Install a container as a Flatpak"""

    installer = Installer(profile=get_profile())
    if koji:
        installer.set_source_koji_name_stream(path_or_url)
    elif path_or_url.startswith("http://") or path_or_url.startswith("https://"):
        installer.set_source_url(path_or_url)
    else:
        installer.set_source_path(path_or_url)

    installer.install()



import koji
import re

FEDORA_TAG_PATTERNS = [(re.compile(p), s) for p, s in [
    (r'^(f\d+)-modular$', 'stable'),
    (r'^(f\d+)-modular-pending$', 'pending'),
    (r'^(f\d+)-modular-signing-pending$', 'signing-pending'),
    (r'^(f\d+)-modular-updates$', 'stable'),
    (r'^(f\d+)-modular-updates-candidate$', 'candidate'),
    (r'^(f\d+)-modular-updates-pending$', 'pending'),
    (r'^(f\d+)-modular-updates-testing$', 'testing'),
]]

STATUSES = sorted({s for p, s in FEDORA_TAG_PATTERNS})


def _add_status_and_base_version(session, build):
    # Find out where a build is tagged to find its base Fedora version and status
    tags = session.listTags(build=build['build_id'])
    for t in tags:
        for p, s in FEDORA_TAG_PATTERNS:
            m = p.match(t['name'])
            if m:
                build['gmb_status'] = s
                build['gmb_base_version'] = m.group(1)
                break


def _add_rpm_list(session, build):
    archives = session.listArchives(buildID=build['build_id'])
    # The RPM list for the 'modulemd.txt' archive has all the RPMs, recent
    # versions of MBS also write upload 'modulemd.<arch>.txt' archives with
    # architecture subsets.
    archives = [a for a in archives if a['filename'] == 'modulemd.txt']
    assert len(archives) == 1
    build['gmb_rpms'] = session.listRPMs(imageID=archives[0]['id'])


def get_module_builds_for_session(session,
                                  module_name, stream,
                                  version=None,
                                  base_version=None,
                                  status=None,
                                  include_rpms=False):

    """Like get_module_builds() but takes and existing Koji session object

    session -- a koji.ClientSession object
    module_name -- the name of the module
    stream -- the stream of the module
    version -- the version of the module. If not specified, the latest
       version will be used.
    base_version -- the base Fedora version that the module was built for
       (corresponds to the stream of the 'platform' module). If None
       builds for all base versions will be returned
    status -- the status of the module in Fedora - can be
       'stable', 'testing', 'candidate', 'pending', or 'signing-ending'. If None,
       builds with all statuses will be returned
    include_rpms -- if %TRUE, each returned build will have a 'gmb_rpms' key with
       the value being a list of dictionaries with information about each RPM
       built in the module.
    """

    package_id = session.getPackageID(module_name)

    # List all builds of the module
    builds = session.listBuilds(packageID=package_id, type='module',
                                state=koji.BUILD_STATES['COMPLETE'])
    # For convenience, add keys that turn the NVR into module terms
    for b in builds:
        b['gmb_stream'] = b['version']
        if '.' in b['release']:
            b['gmb_version'], b['gmb_context'] = b['release'].split('.', 1)
        else:
            b['gmb_version'], b['gmb_context'] = b['release'], None

        # Will fill in later
        b['gmb_status'] = None
        b['gmb_base_version'] = None

    def filter_build(b):
        if b['gmb_stream'] != stream:
            return False
        if version is not None and b['gmb_version'] != version:
            return False
        return True

    matching_builds = [b for b in builds if filter_build(b)]

    if base_version is not None or status is not None:
        for b in matching_builds:
            _add_status_and_base_version(session, b)

        def filter_build_2(b):
            if base_version is not None and b['gmb_base_version'] != base_version:
                return False
            if status is not None and b['gmb_status'] != status:
                return False
            return True

        matching_builds = [b for b in matching_builds if filter_build_2(b)]

    if version is None and len(matching_builds) > 0:
        # OK, we've limited the builds to the ones that match the search criteria, find
        # the most recent one, based on the module version.
        latest_version = max(b['gmb_version'] for b in matching_builds)
        result = [b for b in matching_builds if b['gmb_version'] == latest_version]
    else:
        result = matching_builds

    # Look up tag information if we didn't do so earlier to filter
    if not (base_version is not None or status is not None):
        for b in result:
            _add_status_and_base_version(session, b)

    if include_rpms:
        for b in result:
            _add_rpm_list(session, b)

    return result


def get_module_builds(module_name, stream,
                      version=None,
                      base_version=None,
                      status=None,
                      koji_config=None,
                      koji_profile='koji',
                      include_rpms=False):

    """Return a list of Koji build objects for the specified, or latest
    version of a module. All the returned builds will have the same version,
    but multiple builds with different contexts may be returned due to
    stream expansion.

    module_name -- the name of the module
    stream -- the stream of the module
    version -- the version of the module. If not specified, the latest
       version will be used.
    base_version -- the base Fedora version that the module was built for
       (corresponds to the stream of the 'platform' module). If None
       builds for all base versions will be returned
    status -- the status of the module in Fedora - can be
       'stable', 'testing', 'candidate', 'pending', or 'signing-ending'. If None,
       builds with all statuses will be returned
    koji_config -- alternate koji config file to read
    koji_profile -- alternate koji profile to use
    include_rpms -- if %TRUE, each returned build will have a 'gmb_rpms' key with
       the value being a list of dictionaries with information about each RPM
       built in the module.
    """

    options = koji.read_config(profile_name=koji_profile, user_config=koji_config)
    session_opts = koji.grab_session_options(options)
    session = koji.ClientSession(options['server'], session_opts)

    return get_module_builds_for_session(session,
                                         module_name, stream,
                                         version=version,
                                         base_version=base_version,
                                         status=status,
                                         include_rpms=include_rpms)



from setuptools import setup
import sys

# Only install the CLI for python3. For python2, all we want to work is
# flatpak_builder.py
if sys.version_info[0] == 3:
    entry_points='''
        [console_scripts]
        flatpak-module=flatpak_module_tools.cli:cli
    '''
else:
    entry_points = None

setup(name='flatpak-module-tools',
      version='0.11.4',
      description='Tools for creating and maintaining Flatpaks as Fedora modules',
      url='https://pagure.io/flatpak-module-tools',
      author='Owen Taylor',
      author_email='otaylor@redhat.com',
      license='MIT',
      packages=['flatpak_module_tools'],
      package_data={'flatpak_module_tools': ['config.yaml', 'templates/*.j2']},
      include_package_data=True,
      entry_points=entry_points,
      )
